package sortmerge;

import java.lang.*;
import java.util.*;
import java.io.*;

class Main
{
  public Main()
  {
  }

  public static void main (String[] args)
  {
    /* Commandline arguements:
     *  0: Input file name
     *  2: maximum number of input files for each merge
     *      (range: 2-15 inclusive)
     */

    if (args.length < 2)
      {
	// We dont have enough parameters; output usage information
	printUsage();
	return;
      }

    // Ok, we got this far so we have enough parameters. Time to pull them
    // out of the array.
    String inputFileName = args[0];
    Integer maxInputFileCount = new Integer(args[1]);
    File inputFile = new File(inputFileName);

    // Check that the maximum input file count is within allowed range:
    if ((maxInputFileCount < 2) || (maxInputFileCount > 15))
    {
      // The maximum input file count is out of range.

      // TODO: Print max IF count out of range error
      printUsage();
      return;
    }

    // Check that the input file exists:

    if (!inputFile.exists())
      {
	// The input file doesnt exist
	// TODO: Print invalid input file error
	printUsage();
	return;
      }

    // Ok, we have managed to get through the checks without returning so
    // everything is OK.

    // Setup the input files:

    // Create temp input files. When the virtual machine terminates nicely
    // the files will be automatically deleted.
    File[] files = new File[maxInputFileCount];
    File[] outfiles = new File[maxInputFileCount];

    for (int i = 0; i < maxInputFileCount;i++)
      {
	try
	  {
	    File tempFile = File.createTempFile("sort",".tmp");
	    tempFile.deleteOnExit();
	    files[i] = tempFile;

	    tempFile = File.createTempFile ("sout", ".tmp");
	    outfiles[i] = tempFile;
	    tempFile.deleteOnExit();

	  }
	catch (Exception e)
	  {
	    // Somthing has gone wrong; temp file not created.
	    e.printStackTrace();
	    return;
	  }
      }

    //Now create the inital runs into files.
    try
      {
	//We open the main input file. This is then read in and has 
	//initial runs created in the "files" array.
	FileInputStream istream = new FileInputStream (inputFile);
	create_initial_runs (istream, 15, files);
	istream.close();

	// Now we have the "files" array and the "outfiles" array. The first
	// should have the runs in them, while the files referred to in the
	// second array should be empty.

        // array [files]    : Currently contains the initial runs
        // array [outfiles] : Should be empty at this time.

        // This variable tells us which are input files and which are
        // output files:
        boolean reverse_in_out = false;
        
       
        // This will store the number of output files used by the merge
        // process.
        int result = files.length;
        
	// Loop until we run out of runs to merge (when result == 1)
	while (result > 1)
	  {
	    // Ok, we have a bunch of runs in tmp_outputs and a bunch of empty 
            // files in tmp_inputs. We must now merge.
	    
	    // Step 1: Merge the runs
	    // The returned value is how many output files were used
	    if (reverse_in_out) result = merge_step (outfiles, files,result);
	    else result = merge_step (files, outfiles,result);

            // Step 2: Swap over tmp_inputs and tmp_outputs
	    if (reverse_in_out) reverse_in_out = false;
	    else reverse_in_out = true;
	  }
        
        // Now we print the final remaining file to standard output
        if (reverse_in_out) printFile(outfiles[0]);
        else printFile(files[0]);

      }
    catch (Exception e)
      {
	e.printStackTrace();
      }
    

  }

  public static void printFile(File theFile) throws Exception {
      FileInputStream istream = new FileInputStream (theFile);
      
      BufferedReader reader = new BufferedReader (new FileReader (istream.getFD()));
      boolean at_eof = false;
      
      while (!at_eof)
      {
        String s = reader.readLine();
        if (s == null)
        {
            // we are at the end of the file.
            at_eof = true;
            break;
        }
        else {
            System.out.println(s);
        }
     }
  }

  /**
   * This function should output some basic usage information.
   */
  public static void printUsage()
  {
    System.err.println ("sortmerge: You should know how to use this");
  }

  /**
   * This function creates the initial runs
   * @param input The input file stream to create runs from
   * @param heapsize The heap size
   * @param files The files to output runs to
   */
  public static void create_initial_runs (FileInputStream input, int heapsize, File[] files) throws Exception
  {
    BufferedReader reader = new BufferedReader (new FileReader (input.getFD()));
    boolean at_eof = false;

    MinHeap heap = new MinHeap (heapsize);
    String last_output = null;
    PrintWriter output = null;

    int file_index = 0;
    output = new PrintWriter (new FileWriter (files[file_index], true));

    while (true)
      {
	//Try and fill up the heap with new input items.
	if (!at_eof)
	  {
	    for (int i = heap.getUsage(); i < heapsize; i++)
	      {
		String s = reader.readLine();

		if (s == null)
		  {
		    //The input stream is at the end of the file.
		    at_eof = true;
		    break;
		  }

		boolean marked_state = false;
		if (last_output != null)
		  {
		    int compare_result = s.compareTo (last_output);
		    System.err.println ("Compare result is " + compare_result);
		    if (compare_result < 0)
		      {
			marked_state = true;
		      }
		  }

		System.err.println ("Adding: " + s + " in state " + marked_state);
		heap.add (new SortInput (s, marked_state));

		assert (heap.checkHeapiness());
	      }
	  }

	//If the string is not marked, then just output
	//it to the current run. (stdout for now).
	SortInput root = heap.getRoot();

	//If the root is null, then we have run out of
	//input altogether.
	if (root == null)
	  {
	    System.err.println ("Root is null; out of input");
	    break;
	  }

	if (!root.getMarkedState())
	  {
	    //Ouput the string and store it as the last
	    //outputted string.
	    output.println (root.getString());
	    last_output = root.getString();

	    //Now read in a new string to take its place.
	    String s = reader.readLine();

	    if (s != null)
	      {
		//Find out whether the value we just read in is smaller than
		//the last output value.
		boolean marked_state = false;
		if (last_output != null)
		  {
		    int compare_result = s.compareTo (last_output);
		    System.err.println ("Compare result is " + compare_result);
		    if (compare_result < 0)
		      {
			marked_state = true;
		      }
		  }

		root.setString (s);
		root.setMarkedState (marked_state);
		heap.downheap();

		assert (heap.checkHeapiness());
	      }
	    else
	      {
		heap.deleteRoot();
		assert (heap.checkHeapiness());
	      }
	    
	  }
	else
	  {
	    output.flush();
	    output.close();
	    
	    file_index++;
	    file_index %= files.length;

	    output = new PrintWriter (new FileWriter(files[file_index], true));

	    heap.markAll (false);
	  }
      }

    output.flush();
    output.close();


  }

  /**
   * This function exists just incase somthing is still using the old merge_step
   * function that doesnt take the additional integer parameter. This function
   * assumes that all input files need merging.
   *
   * @param inputs The array of input files to merge
   * @param outputs The array of output files to merge into
   * @returns The number of output files used
   */
  public static int merge_step (File[] inputs, File[] outputs) throws Exception {
      return merge_step(inputs,outputs,inputs.length);
  }
  
  /**
   * This all new merge_step takes an integer parameter stating how many of the
   * elements in the inputs array need merging.
   *
   * @param inputs The array of input files to merge
   * @param outputs The array of output files to merge into
   * @param incount The number of elements in the input files array that need merging
   * @returns The number of output files used
   */
  public static int merge_step (File[] inputs, File[] outputs, int incount) throws Exception
  {
    //We are given a list of input files to merge, and a list of files
    //we can use for output.
    
    PrintWriter output = null;
    int output_index = 0;
    String last_output = null;
    
    //Create a new heap which will be used to select the
    //file with the next smallest string.
    MinHeap heap = new MinHeap (inputs.length);

    //Populate the heap initially.
    //for (int i = 0; i < inputs.length; i++)
    for (int i = 0; i < incount; i++)
      {
	try
	  {
	    heap.add (new SortInput (new FileInputStream (inputs[i])));
	  }
	catch (EOFException e)
	  {
	    //We had an empty input file, nevermind.
	    continue;
	  }
      }

    assert (heap.checkHeapiness());

    boolean[] do_append = new boolean[outputs.length];

    output = new PrintWriter (new FileWriter (outputs[output_index], 
					      do_append [output_index]));


    while (true)
      {
	SortInput root = heap.getRoot();

	if (root == null)
	  {
	    //There is nothing left in our heap; this means
	    //that we have got to the end of our input files.
	    break;
	  }

	if (!root.getMarkedState())
	  {
	    output.println (root.getString());
	    last_output = root.getString();
	    
	    try
	      {
		root.next();
		
		String s = root.getString();
		
		//Now check the value we just read out against
		//the last item we output. If it is smaller, it needs
		//to go into another run.
		boolean marked_state = false;
		int compare_result = s.compareTo (last_output);
		
		if (compare_result < 0)
		  {
		    marked_state = true;
		  }
		
		root.setMarkedState (marked_state);
		
		heap.downheap();
		assert (heap.checkHeapiness());
	      }
	    catch (EOFException e)
	      {
		heap.deleteRoot();
		assert (heap.checkHeapiness());
	      }
	    
	  }
	else
	  {
	    //Really we should swap outputs.
	    output.flush();
	    output.close();

	    do_append[output_index] = true;

	    output_index++;
	    output_index %= outputs.length;

	    output = new PrintWriter (new FileWriter (outputs[output_index],
						      do_append[output_index]));

	    heap.markAll (false);
	  }
      }

    output.flush();
    output.close();
    
    //Count the number of output files that were used.
    int used_count;
    for (used_count = 0; used_count < do_append.length; used_count++)
      {
	if (!do_append[used_count])
	  {
	    break;
	  }
      }

    //If there was only one file used; then that count would
    //have returned zero... but thats wrong.
    if (used_count == 0)
      {
	used_count = 1;
      }

    return used_count;
  }
}
				      
