package sortmerge;

import java.util.*;
import java.io.*;
import java.lang.*;

class MinHeap
{
  protected SortInput[] heap = null;
  protected int next_node = 0;

  /**
   * Construct a minheap with the specified size.
   * @param size The size of the heap to create.
   */
  public MinHeap (int size)
  {
    heap = new SortInput[size];
    for (int i = 0; i < size; i++)
      {
	heap[i] = null;
      }
  }

  /**
   * Add an item into the heap.
   * @param item Item to add into the heap.
   */
  public void add (SortInput item) throws Exception
  {
    if (next_node >= heap.length)
      {
	throw new Exception ("Heap is full");
      }

    //System.err.println ("---------- Beginning add operation ----------");
    //Add the item into the bottom left most free
    //node in the heap.
    heap[next_node] = item;

    //We will now upheap, starting at the new node
    //which will ensure the heap property is not violated.
    upheap (next_node);

    //Increase the next node 'pointer'
    next_node++;
    //System.err.println ("------------ End add operation ----------");
  }

  /**
   * Upheap the heap, starting at the given node. This
   * makes sure the heap property is maintained.
   * @param node The node to upheap from.
   */
  protected void upheap (int node)
  {
    if (node == 0)
      {
	//We have got back to the root.
	//System.err.println ("Node was the parent; stopping upheap");
	return;
      }

    int parent_index = getParent (node);
    //System.err.println ("Current index is " + node + " - Parent index was " + parent_index);
    
    SortInput parent = heap[parent_index];
    SortInput item = heap[node];

    //Check that the heap property is maintained.
    int compare_result = item.compareTo (parent);

    //System.err.println ("Compare result was " + compare_result);

    //Because this is a minheap, the current item should
    //be less than.
    if (compare_result < 0)
      {
	//System.err.println ("Current item was less than parent; swap then upheap");
	heap[node] = parent;
	heap[parent_index] = item;
	
	upheap (parent_index);
      }
  }

  /**
   * Get the root item from the heap. This is the smallest one.
   * @return The root item, or null if empty.
   */
  public SortInput getRoot()
  {
    return heap[0];
  }

  public void deleteRoot()
  {
    //System.err.println ("--------- Beginning delete ---------");

    //If the next node is 0, then do nothing -- the heap is empty.
    if (next_node == 0)
      {
	//System.err.println ("The heap is empty.");
	return;
      }

    //The only node in the heap is the root node; so we will just 
    //replace it with null.
    if (next_node == 1)
      {
	//System.err.println ("Only root node exists, replacing with null.");
	heap[0] = null;
	next_node--;
	return;
      }

    //Swap the last item in the heap with the root.
    heap[0] = heap[next_node - 1];
    
    //Reduce the next-node 'pointer'
    next_node--;

    //Now downheap starting at the root to ensure the
    //heap property still applies.
    downheap (0);
    
    //System.err.println ("--------- End delete ----------");
  }

  /**
   * Downheap the heap, starting at the root node.
   */
  public void downheap()
  {
    downheap (0);
  }

  /**
   * Downheap, starting at the specified node.
   * @param node The node to begin downheaping at.
   */
  protected void downheap (int node)
  {

    int left_index = getLeft (node);
    int right_index = getRight (node);

    if (left_index >= next_node && right_index >= next_node)
      {
	//Both child nodes are out of range; so we are at the
	//bottom of the heap.
	//System.err.println ("Both child indexes were out of range, at leaf node");
	return;
      }

    //If just the left node is within the bounds we still
    //have enough data to have to swap down again.
    int smaller_index = left_index;
    int compare_result = -1;

    //If both children exist, then we will find the larger one.
    if (left_index < next_node && right_index < next_node)
      {
	//System.err.println ("Node: " + node + "; Left index: " + left_index + "; Right index: " + right_index);
	
	//Compare the left and right children. We
	//use this to decide which branch we downheap into.
	compare_result = heap[left_index].compareTo (heap[right_index]);
	
	//System.err.println ("Compare result: " + compare_result + "; Smaller index: " + smaller_index);
	
	
	if (compare_result >= 0)
	  {
	    //The right child was the smaller branch.
	    //System.err.println ("Right child was smaller");
	    smaller_index = right_index;
	  }
      }
	
    compare_result = heap[node].compareTo(heap[smaller_index]);
    //System.err.println ("Second compare result: " + compare_result);
	
    if (compare_result >= 0)
      {
	//Swap the current node with the smaller index.
	//System.err.println ("Swapping down and downheap");
	SortInput smaller = heap[smaller_index];
	heap[smaller_index] = heap[node];
	heap[node] = smaller;

	//Now downheap from the swapped node.
	downheap (smaller_index);
      }
  }

  /**
   * Mark all values in the heap to a given state.
   * @param mark_state the value to set the mark to on each element.
   */
  public void markAll(boolean mark_state)
  {
    for (int i = 0; i < next_node; i++)
      {
	heap[i].setMarkedState (mark_state);
      }
  }

  /**
   * Get the count of items in the current heap.
   * @return the number of items in the heap.
   */
  public int getUsage()
  {
    return next_node;
  }

  /**
   * Get the index for the left child of heap node.
   * @param The index of the node to get the child of.
   * @return The index of the left child node.
   */
  protected int getLeft (int node)
  {
    return (2 * node) + 1;
  }

  /**
   * Get the index for the right child of the heap node.
   * @param The index of the node to get the child of.
   * @return The index of the right child node.
   */
  protected int getRight (int node)
  {
    return (2 * node) + 2;
  }

  /**
   * Get the index of the parent node. If the node is
   * the root node, then the result will be less than 0.
   * @param node The node to get the parent of.
   * @return The index of the parent node.
   */
  protected int getParent (int node)
  {
    return (node - 1) / 2;
  }

  /**
   * Ensure the heap property for this MinHeap hasn't been 
   * violated.
   * @return True if the heap is valid.
   */
  public boolean checkHeapiness()
  {
    //Go backwards through the heap and check the parents are
    //smaller than the node.
    for (int i = next_node - 1; i > 0; i--)
      {
	int parent = getParent (i);

	if (heap[parent].compareTo (heap[i]) > 0)
	  {
	    return false;
	  }
      }
    
    return true;
  }
  
}
